from munch import Munch

from infi.shell import (
    commandgroup,
    command,
    )
from infi.shell.exceptions import BadValue
from infi.shell.types import YesNo
from infi.shell.utils import strlist

from ..exceptions import ManagementException
from ..renderer import render_capacity
from .hierarchical_base import HierarchicalEntityCommandsBase
from .types import (
    Pool,
    Filesystem,
    FilesystemSnap,
    FilesystemClone,
    FilesystemOrClone,
    MasterFilesystem,
    Name,
    Size,
    SizeDelta,
    CapacityUnit,
    FSSeries,
    )
from .options import QueryOptions
from .utils import get_target_pool
from . import render_collection, roles, set_rendered_unit, set_renderer_query_fields

FS_QUERY_FIELDS                     = strlist("name                        thin size used allocated tree_allocated pool write_protected mapped")
DETAILED_FS_QUERY_FIELDS            = strlist("name nfs_filesystem_id      thin size used allocated tree_allocated pool write_protected mapped ssd_enabled root_mode created_at updated_at")

SNAP_QUERY_FIELDS                   = strlist("name                        thin size used allocated                pool                 mapped                       created_at")
DETAILED_SNAP_QUERY_FIELDS          = strlist("name nfs_filesystem_id      thin size used allocated                pool                 mapped ssd_enabled root_mode created_at updated_at")

CLONE_QUERY_FIELDS                  = strlist("name                        thin size used allocated tree_allocated pool write_protected mapped                       created_at")
DETAILED_CLONE_QUERY_FIELDS         = DETAILED_FS_QUERY_FIELDS

FS_TREE_FIELDS                      = strlist("name                   type      size used allocated                pool write_protected                              created_at")
DETAILED_FS_TREE_FIELDS             = strlist("name nfs_filesystem_id type thin size used allocated tree_allocated pool write_protected mapped ssd_enabled root_mode created_at updated_at")

SORTABLE_FIELDS = strlist("name serial size write_protected mapped ssd_enabled created_at updated_at")

QUERY_SUBSTITUTES = dict(
    write_protected="wp",
    ssd_enabled="ssd_cache",
    root_mode="default_filesystem_permissions",
    nfs_filesystem_id="nfs_fs_id",
    mapped="exported",
    )

DETAILED_QUERY_SUBSTITUTES = dict(
    write_protected="write_protected",
    )

class FilesystemCommandsMixin(object):
    FS_TYPE_NAMES = Munch(
        MASTER="filesystem",
        SNAP="snapshot",
        CLONE="clone",
        )

    def _get_fs_type(self, fs):
        if fs.type == "MASTER":
            return "FILESYSTEM"
        return fs.type

    def _get_fs_type_name(self, volume):
        return self.FS_TYPE_NAMES.get(volume.type) or "filesystem"

class FilesystemCommandsBase(FilesystemCommandsMixin, HierarchicalEntityCommandsBase):
    HIERARCHICAL_TYPE = Filesystem

    def _query_hook(self, filesystems, pool_names_by_id=None):
        for fs in filesystems:
            if pool_names_by_id:
                fs.pool = pool_names_by_id[fs.pool_id]
            fs.type = self._get_fs_type(fs)
            fs.thin = fs.provtype == self.THIN
        return filesystems

def create_common_filesystem_commands_mixin(FilesystemType, fs_arg):

    class CommonFilesystemCommandsMixin(object):

        @command(
            description="Rename {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, Name),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, fs, new_name):
            return self._rename(_api, fs, new_name)

        @command(
            description="Delete {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType.List,),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, fs):
            return self._delete(_api, fs)

        @command(
            description="Enable or disable SSD cache",
            types=(FilesystemType, YesNo, YesNo),
            substitute=dict(fs="name", enabled="ssd_cache"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def cache(self, _api, fs, enabled, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="{} cached on SSD".format("now" if enabled else "no longer"),
                success_msg_with_is=True,
                noop_msg="already{} cached on SSD".format("" if enabled else " not"),
                noop_msg_with_is=True,
                props=dict(
                    ssd_enabled=enabled,
                    ),
                predicate=lambda dataset: dataset.ssd_enabled != enabled,
                )

    return CommonFilesystemCommandsMixin

def create_writable_filesystem_commands_mixin(FilesystemType, fs_arg):

    class WritableFilesystemCommandsMixin(object):

        @command(
            description="Resize {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, SizeDelta),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def resize(self, _api, fs, size):
            size = size.compute_size(fs.size)
            if fs.size == size:
                return self._result_message(fs, "size is already {}".format(render_capacity(fs.size)))
            self._update(_api, fs, size=size)
            return self._result_message(fs, "resized")

        @command(
            description="Restore data state of {} from previous snapshot".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, FilesystemSnap),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def restore(self, _api, fs, source):
            if fs.id == source.id:
                raise BadValue('"{}" was specified as both arguments'.format(fs.name))
            _api.post(FilesystemType.get_uri(fs.id, "restore"), source.id)
            # self._post(_api, fs, "restore", data=dict(
            #     source_id=source.id
            #     ))
            return self._result_message(fs, 'restored from {} "{}"'.format(self._get_fs_type_name(source), source.name))

        @command(
            description="Disable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_protect(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-protected",
                success_msg_with_is=True,
                noop_msg="already write-protected",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=True,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and not dataset.write_protected,
                )

        @command(
            description="Enable write access on {}".format(FilesystemType.DESCRIPTION_SUBJECT),
            types=(FilesystemType, YesNo),
            substitute=dict(fs="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_enable(self, _api, fs, recursive=False):
            return self._update_recursive(_api,
                entity=fs,
                recurse=recursive,
                success_msg="now write-enabled",
                success_msg_with_is=True,
                noop_msg="already write-enabled",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=False,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and dataset.write_protected,
                )

    return WritableFilesystemCommandsMixin

def create_child_filesystem_commands_mixin(FilesystemType, ParentType, fs_arg, parent_arg, create_parent_arg):

    class ChildFilesystemCommandsMixin(object):

        @command(
            description="Create a new {} from {}".format(FilesystemType.DESCRIPTION_SUBJECT, ParentType.DESCRIPTION_SUBJECT),
            types=(ParentType, Name, YesNo),
            substitute=dict(fs=create_parent_arg),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def create(self, _api, fs, name, ssd_cache=None):
            parent = self._create_and_return_entity(_api, parent_id=fs.id, name=name, ssd_enabled=ssd_cache)
            return self._result_message(parent, "created", subject=FilesystemType.get_entity_name())

        @command(
            description="Display hierarchy of snapshots and clones",
            types=(FilesystemType, CapacityUnit),
            substitute=dict(fs="name"),
            options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Filesystem,
            fields=FS_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
        )
        def tree(self, _api, fs, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, fs)

    return ChildFilesystemCommandsMixin

@commandgroup(prefix="fs")
class FilesystemCommands(FilesystemCommandsBase,
                         create_common_filesystem_commands_mixin(MasterFilesystem, "fs"),
                         create_writable_filesystem_commands_mixin(MasterFilesystem, "fs")):
    ENTITY_TYPE = MasterFilesystem
    @command(
        description="List existing filesystems",
        types=(MasterFilesystem.List, Pool, CapacityUnit, YesNo, YesNo, YesNo),
        substitute=dict(fs="name"),
        options=QueryOptions(fields=DETAILED_FS_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_QUERY_FIELDS,
        detailed_fields=DETAILED_FS_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, fs=None, pool=None, unit=None, has_exports=None, write_protected=None, thin=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, fs, pool_id=pool, mapped=has_exports, write_protected=write_protected, provtype=provtype)

    @command(
        description="Create a new filesystem",
        types=(Name.List, Size, Pool, YesNo, YesNo, FSSeries),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, series=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple filesystems, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # filesystems.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self.THIN if thin else self.THICK,
            write_protected=False,
            ssd_enabled=ssd_cache,
            )

    @command(
        description="Move filesystem to a different pool",
        types=(MasterFilesystem, Pool, YesNo),
        substitute=dict(fs="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, fs, pool, move_capacity=False):
        if fs.pool_id == pool.id:
            return self._result_message(fs, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        _api.post(self.ENTITY_TYPE.get_uri(fs.id, "move"), dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(fs, 'moved to pool "{}" {} its capacity'
                                        .format(pool.name, "along with" if move_capacity else "without"))

    @command(
        description="Set filesystem provisioning type",
        types=(MasterFilesystem, YesNo),
        substitute=dict(fs="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thin(self, _api, fs, thin):
        prov_name = "thin" if thin else "thick"
        # if fs.provtype.upper() == self._get_provtype(thin):
        #     raise CommandAborted(self._result_message(fs, "already set to {} provisioning".format(prov_name)))
        self._update(_api, fs, provtype=self._get_provtype(thin))
        return self._result_message(fs, "set to {} provisioning".format(prov_name))

    @command(
        description="Estimate reclaimed capacity (deletion simulation)",
        types=(Filesystem,),
        substitute=dict(filesystem="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def reclaimable(self, _api, filesystem):
        data = dict(entities=[filesystem.id])
        result = _api.post(Filesystem.get_uri("delete_simulation"), data)
        return 'Estimated reclaimable space for filesystem "{}" and its tree descendants: {}'.format(filesystem.name, render_capacity(result.space_reclaimable))

    @command(
        description="Display hierarchy of filesystems, snapshots and clones",
        types=(MasterFilesystem, Pool, CapacityUnit),
        substitute=dict(fs="name"),
        options=QueryOptions(fields=DETAILED_FS_TREE_FIELDS, sortable=FS_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Filesystem,
        fields=FS_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def tree(self, _api, fs=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, fs, pool)

@commandgroup(prefix="fs.snap")
class FilesystemSnapshotCommands(FilesystemCommandsBase,
                                 create_common_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnap,
                                     fs_arg="snap",
                                     ),
                                 create_child_filesystem_commands_mixin(
                                     FilesystemType=FilesystemSnap,
                                     ParentType=FilesystemOrClone,
                                     fs_arg="snap",
                                     parent_arg="fs",
                                     create_parent_arg="source",
                                     )):
    ENTITY_TYPE = FilesystemSnap

    @command(
        description="List existing snapshots",
        types=(FilesystemSnap.List, FilesystemOrClone, Pool, CapacityUnit, YesNo),
        substitute=dict(snap="name"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=SNAP_QUERY_FIELDS,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, snap=None, source=None, pool=None, unit=None, has_exports=None):
        set_rendered_unit(unit)
        return self._common_query(_api, snap, parent_id=source, pool_id=pool, type=FilesystemSnap.ENTITY_MGMT_SUBTYPE, mapped=has_exports)

@commandgroup(prefix="fs.clone")
class FilesystemCloneCommands(FilesystemCommandsBase,
                              create_common_filesystem_commands_mixin(
                                  FilesystemType=FilesystemClone,
                                  fs_arg="clone",
                                  ),
                              create_writable_filesystem_commands_mixin(
                                  FilesystemType=FilesystemClone,
                                  fs_arg="clone",
                                  ),
                              create_child_filesystem_commands_mixin(
                                  FilesystemType=FilesystemClone,
                                  ParentType=FilesystemSnap,
                                  fs_arg="clone",
                                  parent_arg="snap",
                                  create_parent_arg="snap",
                                  )):
    ENTITY_TYPE = FilesystemClone

    @command(
        description="List existing clones",
        types=(FilesystemClone.List, FilesystemSnap, Pool, CapacityUnit, YesNo, YesNo),
        substitute=dict(clone="name"),
        options=QueryOptions(fields=DETAILED_CLONE_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Filesystem,
        fields=CLONE_QUERY_FIELDS,
        detailed_fields=DETAILED_CLONE_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, clone=None, snap=None, pool=None, unit=None, has_exports=None, write_protected=None):
        set_rendered_unit(unit)
        return self._common_query(_api, clone, parent_id=snap, pool_id=pool, type=FilesystemClone.ENTITY_MGMT_SUBTYPE, mapped=has_exports, write_protected=write_protected)
