import re
from contextlib import contextmanager

from infi.shell import runtime
from infi.shell.config import config, altered_config
from infi.shell.exceptions import BadArguments

from ...utils import input_password
from ..system.system import api
from ..system.user import Roles
from ..exceptions import InvalidSystemState
from .types import (
    Pool,
    PatchPanelFrame,
    )

PRIVILEGED_ROLES = set((Roles.INFINIDAT, Roles.ADMIN))

def is_privileged_user(user):
    """ Check whether or not a user entity is an Infinidat or admin user """
    return any(r.upper() in PRIVILEGED_ROLES for r in user.roles)

def user_has_role(user, role):
    """ Check whether or not a user entity has the specified role """
    return role.upper() in (r.upper() for r in user.roles)

def is_pool_admin(user):
    """ Check whether or not a user entity is a pool admin """
    return user_has_role(user, Roles.POOLADMIN)

@contextmanager
def increased_api_timeout(timeout_seconds=300):
    """ Temporarily increase API timeout to 5 minutes, or to the specified
    value. If the current API timeout is greater than 5 minutes or the specified
    value, nothing is changed. """
    if (config.api_request_timeout_seconds or 0) > timeout_seconds:
        yield
    else:
        with altered_config(api_request_timeout_seconds=timeout_seconds):
            yield

def are_none(*values):
    return all(v is None for v in values)

def any_is_none(*values):
    return any(v is None for v in values)

def is_detailed_query():
    return runtime.options.get("detailed", False)

def get_target_pool(_api, pool=None):
    """
    Get target pool for creating a new storage entity (volume, fs, etc'). If
    the specified pool is not ``None``, it is returned. Otherwise, if only
    one pool exists in the system, then it is returned. Otherwise, a
    ``BadArguments`` exception is raised.
    """
    if pool:
        return pool
    pools = Pool.paged_query(_api, page_size=1)
    if pools.total_objects == 0:
        raise InvalidSystemState("There are no pools in the system, please create a pool first")
    if pools.total_objects > 1:
        raise BadArguments("There are multiple pools in the system, please specify a pool")
    return next(iter(pools))

_PORT_NAME_PATTERN = re.compile(r"^\s*\w+(\d+)\w+(\d+)\s*$")
def get_patch_panel_ports(port_type):
    """
    Get a dict where each key is a 2-tuple of (node_id, physical_port_id) and
    each value is a patch-panel port object.
    :param port_type: Port type, e.g "FC" or "ETH"
    """
    port_type = port_type.upper()
    result = {}
    for frame in PatchPanelFrame.iter_all(api):
        for pp_port in frame.ports:
            if not (pp_port.conn_port_comp_id and pp_port.type and pp_port.type.upper() == port_type):
                continue
            match = _PORT_NAME_PATTERN.match(pp_port.conn_port_comp_id)
            if not match or match.lastindex != 2:
                continue
            try:
                result[(int(match.group(1)), int(match.group(2)))] = pp_port
            except ValueError:
                continue
    return result

def input_password_check_credentials(prompt=None, username=None, interaction_error=None):
    """ Input a user's password, and validate the credentials. When ``username``
    is ``None``, use the current logged-in user to validate the credentials. """
    from .system import system
    return system.check_credentials(username or system.user.username, input_password(prompt, interaction_error=interaction_error))

def strip_newline_chars(s):
    """ Replace newline characters in a string. """
    if s:
        s = s.replace("\n", "; ")
    return s


# The following function's implementation depends on the OS.
from sys import platform

if platform == 'darwin' or platform.startswith('linux'):
    # In OS X and Linux we can wait for keystrokes with select() on sys.stdin.
    # (Windows doesn't support select() on files, so this code will not work.)
    import select
    import sys
    import termios
    import time
    import tty

    CONTROL_C = ('\x03', 'Q', 'q')

    def getch_pause(timeout=5.0):
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        t_now = time.time()
        t_end = t_now + timeout
        try:
            tty.setraw(fd)
            while t_end > t_now:
                inputs, _, _ = select.select([sys.stdin], [], [], t_end - t_now)
                if sys.stdin in inputs:
                    c = sys.stdin.read(1)
                    if c in CONTROL_C:
                        raise KeyboardInterrupt()
                    t_now = time.time()
                else:
                    break
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)
            termios.tcflush(fd, termios.TCIOFLUSH)

    getch_pause.prefer_Q = True

elif platform.startswith('win'):
    # Windows has a magical console API of its own: kbhit() and getwch().
    import msvcrt
    import time

    CONTROL_C = ('Q', 'q')
    SPECIAL = ('\000', '\xe0')

    def getch_pause(timeout=5.0):
        t_now = time.time()
        t_end = t_now + timeout

        while t_end > t_now:
            if msvcrt.kbhit():
                c = msvcrt.getwch()
                if c in CONTROL_C:
                    raise KeyboardInterrupt()
                elif c in SPECIAL:
                    # Discard the special key code that follows
                    msvcrt.getwch()
            else:
                time.sleep(0.1)
            t_now = time.time()

    getch_pause.prefer_Q = True

else:
    # By default (in other OS) we call sleep instead
    from time import sleep as getch_pause
