from __future__ import absolute_import

import json
from itertools import chain
from numbers import Number
from collections import OrderedDict
from infi.shell.shell import get_active_shell

from .renderer import RendererBase
from .values import ValueRenderer

class JSONRenderer(RendererBase):
    DEFAULT_SEPARATOR = ","
    INDENTATION = 4
    def __init__(self, *args, **kwargs):
        super(JSONRenderer, self).__init__(*args, **kwargs)
        self._value_renderer = ValueRenderer(empty_value="")
    def _write_json(self, obj):
        self._write(json.dumps(obj, indent=self.INDENTATION))
    def _render_impl(self, obj):
        command_text = get_active_shell().current_command.text
        result = {}
        result['command'] = command_text
        if obj is None or isinstance(obj, (basestring, Number)):
            result['result']=obj
            self._write_json(result)
            self._writeln()
            return
        elif self._properties.get("individually", False) and isinstance(obj, dict):
            obj = [obj]
        self._writeln("{")
        self._writeln("\"command\": {},".format(json.dumps(command_text)))
        self._writeln("\"result\": [")
        self._render_collection(obj)
        self._writeln("\n]\n}")
    def _render_collection(self, objects):
        it = iter(objects)
        try:
            first_object = next(it)
        except StopIteration:
            self._write("{}")
            return
        field_names, rendered_field_names, _ = self._get_ordered_and_rendered_fields(first_object)
        all_objects = chain([first_object], it)
        # Rows
        rendered_objects = (self._render_object(o, field_names) for o in all_objects)
        with self._indent(0):  # TODO: Indentation is broken for some reason - please fix
            for i, obj in enumerate(sorted(rendered_objects)):
                if i != 0:
                    self._writeln(",")
                self._write_json(obj)
    def _render_object(self, obj, field_names):
        obj_dict =  {field : self._render_value(obj, field).encode('utf-8', 'ignore') for field in field_names}
        return OrderedDict(sorted(obj_dict.items()))
    def _render_value(self, obj, field):
        value = obj.get(field, self.ERROR_MISSING_FIELD)
        if value is not self.ERROR_MISSING_FIELD:
            value = self._value_renderer.render(field, value, **self._properties.get("value_properties", {}))
        return value

def toggle_json_mode(shell=None):
    shell = shell or get_active_shell()
    if not is_json_mode(shell):
        set_json_mode(True, shell)
        return True
    set_json_mode(False, shell)
    return False

def is_json_mode(shell=None):
    return isinstance((shell or get_active_shell()).get_renderer_override(), JSONRenderer)

def set_json_mode(enabled, shell=None):
    shell = shell or get_active_shell()
    if enabled:
        shell.set_renderer_override(JSONRenderer())
        shell.disable_stream_encoding()
    else:
        shell.set_renderer_override(None)
        shell.enable_stream_encoding()