from functools import partial
from collections import OrderedDict

from infi.shell import (
    config as shell_config,  # Avoid name collision with the "config" commands module of this package
    runtime,
    invalidates_autocompletion_cache,
    availability,
    render,
    get_active_shell,
    )
from infi.shell.commands.commands import metadata as metadata_decorator
from infi.shell.exceptions import BadArguments
from infi.shell.utils import NOTHING, Lazy, DictProxy
from infi.shell.renderer import set_renderer_properties, get_default_renderer_properties

from ...utils import NamedDefaultDict
from ..renderer import CollectionRenderer
from ..system import system as current_system  # There's a module named "system"
from ..system.user import Roles
from .utils import is_detailed_query
from .types import Entity

Null, MESSAGE, ENTITY = object(), object(), object()

class EntityCommands(object):
    ENTITY_TYPE = None

    def __init__(self):
        assert self.ENTITY_TYPE

    def _get_actual_property_value(self, v):
        if v is Null:
            return None
        if isinstance(v, dict) and "id" in v:
            return v["id"]
        return v

    def _get_filtered_property_value(self, value):
        if isinstance(value, (tuple, list, set)):
            return tuple(self._get_actual_property_value(item) for item in value)
        return self._get_actual_property_value(value)

    def _filter_properties(self, properties):
        """ Returns a dict of property names and their values. Values which are
        ``Null`` (see constant) converted to ``None``, and entities are
        converted to plain IDs. """
        return {property : self._get_filtered_property_value(value)
                for property, value in properties.iteritems() if value is not None}
        return dict(self._iter_properties(properties))

    def _filter_entities(self, entities, **conditions):
        """ Get only those entities that match the specified conditions. """
        return EntityFilter(**self._filter_properties(conditions)).filter(entities)

    def _create_query_filter(self, _api, properties):
        """ Filter properties using _filter_properties() and create and return
        a query filter according to them. """
        filter = _api.create_filter()
        for field, value in self._filter_properties(properties).iteritems():
            if isinstance(value, (tuple, list, set, dict)):
                filter = filter.member_of(field, *value)
            else:
                filter = filter.eq(field, value)
        return filter

    def _query(self, _api, _entity=None, _hook=None, _default_sort=None, _sort_on_client_side=False, **conditions):
        """ Get iterator for all entities matching the specified conditions.
        ``_entity`` may be one entity or a list/set/tuple of entities, and when
        specified, the entities are returned, providing they match the
        specified conditions, if any. """
        set_renderer_query_fields()
        sort_fields = runtime.options.get("sort", None)
        sort = self._create_sort(_api, sort_fields) if sort_fields else _default_sort
        entities = None
        if _entity:
            # Entity was specified - use it
            entities = _entity if isinstance(_entity, (list, tuple, set)) else [_entity]
            entities = self._filter_entities(entities, **conditions)
            if not entities:
                return None
            if _hook:
                entities = _hook(entities)
            if sort:
                sort.sort_entities(entities)
            return entities
        query = partial(
            self.ENTITY_TYPE.iter_all,
            _api,
            hook=_hook,
            sort=sort if not _sort_on_client_side else None,
            )
        if conditions and not self.ENTITY_TYPE.FILTERABLE:
            entities = self._filter_entities(query(), **conditions)
        else:
            entities = query(filter=self._create_query_filter(_api, conditions))
        if _sort_on_client_side and sort:
            entities = sort.sorted_entities(entities)
        return entities

    def _create_sort(self, _api, sort_fields):
        sort = _api.create_sort()
        for field, is_ascending in sort_fields or ():
            sort = sort.sort(field, is_ascending)
        return sort

    def _paged_query(self, _api, *args, **kwargs):
        """ Get a PagedQuery object for this group's entity type. """
        assert self.ENTITY_TYPE.PAGED, '{} does not support paging'.format(self.ENTITY_TYPE.get_entity_name())
        return self.ENTITY_TYPE.paged_query(_api, *args, **kwargs)

    def _matches_query(self, entity, conditions):
        for property, value in conditions.iteritems():
            if not hasattr(entity, property) or not value == getattr(entity, property):
                return False
        return True

    def _result_message(self, entity, msg, subject=None):
        entity_type = None
        if isinstance(subject, type) and issubclass(subject, Entity):
            entity_type, subject = subject, None
        return '{} {}'.format(self._result_subject_and_entity(entity, entity_type, subject, capitalize=True), msg)

    def _result_subject_and_entity(self, entity, entity_type=None, subject=None, capitalize=False):
        return '{} {}'.format(
            self._result_subject(subject, entity_type, capitalize=capitalize),
            self._result_entity_name(entity, entity_type),
            )

    def _result_subject(self, subject=None, entity_type=None, capitalize=False):
        entity_type = entity_type or self.ENTITY_TYPE
        result = subject or entity_type.get_entity_name()
        return result if not capitalize else (result[0].upper() + result[1:])

    def _result_entity_name(self, entity, entity_type=None):
        entity_type = entity_type or self.ENTITY_TYPE
        return '"{}"'.format(entity[entity_type.NAME_FIELD])

    def _multiple_results_message(self, operation, entities_or_names, entity_type=None):
        entity_type = entity_type or self.ENTITY_TYPE
        names = (thing if isinstance(thing, basestring) else thing[entity_type.NAME_FIELD] for thing in entities_or_names)
        return "{} {}: {}".format(operation, entity_type.get_entity_plural_name(), ", ".join('"{}"'.format(name) for name in names))

    @invalidates_autocompletion_cache
    def _create(self, _api, returns=MESSAGE, **properties):
        """ Create an entity and return an appropriate message. Properties of which value is None are ignored.
        Properties of which value is Null (see Null constant) are sent as null. """
        entity = _api.post(self.ENTITY_TYPE.get_uri(), self._filter_properties(properties))
        if _api.is_aggregating():
            # Result is None when aggregating, since request is sent only later
            return
        if returns is ENTITY:
            return entity
        if returns is MESSAGE:
            return self._result_message(entity, "created")
        return returns

    def _create_and_return_entity(self, _api, **properties):
        return self._create(_api, returns=ENTITY, **properties)

    def _create_single_or_multiple_or_series(self, _api, names, series, **props):
        if series:
            if len(names) > 1:
                raise BadArguments("Multiple names are not allowed when creating a series")
            return self._create_series(_api, names[0], series, **props)
        if len(names) > 1:
            return self._create_multiple(_api, names, **props)
        return self._create(_api, name=names[0], **props)

    def _create_multiple(self, _api, names, **props):
        with _api.aggregation() as aggregator:
            for name in names:
                aggregator.next_task(name, "Created")
                self._create(_api, name=name, **props)
        return self._multiple_results_message("Created", names)

    def _create_series(self, _api, base_name, series, **props):
        padding = None
        if series.start > 1:
            padding = self._get_series_anchor(_api, base_name, series.start - 1)
        if padding is None:
            padding = self._get_series_anchor(_api, base_name, series.end + 1)
        name_generator = series.generator(base_name, padding)
        if series.total == 1:
            return self._create(_api, name=next(name_generator), **props)
        first_name, last_name = None, None
        with _api.aggregation() as aggregator:
            for name in name_generator:
                if first_name is None:
                    first_name = name
                last_name = name
                aggregator.next_task(name, "Created")
                self._create(_api, name=name, **props)
        return 'Created {count} {entities}: "{first}" through "{last}"'.format(
            count=series.total,
            entities=self.ENTITY_TYPE.get_entity_plural_name(),
            first=first_name,
            last=last_name,
            )

    def _get_series_anchor(self, _api, base_name, anchor):
        series_anchor_candidates = [(base_name + "0"*n + str(anchor)) for n in xrange(1, 7)]
        series_anchor_query = self._paged_query(_api, filter=_api.create_filter().member_of("name", *series_anchor_candidates), page_size=1)
        if series_anchor_query.total_objects == 0:
            return None
        candidate_name = series_anchor_query.get_page(1).objects[0].name
        return len(candidate_name[len(base_name):])

    def _update(self, _api, entity, returns=MESSAGE, field_update_order=(), **properties):
        """ Update an entity and return an appropriate message. Properties of which value is None are ignored.
        Properties of which value is Null (see Null constant) are sent as null. """
        properties = self._filter_properties(properties)
        if self.ENTITY_TYPE.MULTI_FIELD_PUT_ALWAYS_ALL_FIELDS:
            for key, value in entity.iteritems():
                if key != self.ENTITY_TYPE.ID_FIELD:
                    properties.setdefault(key, value)
        if self.ENTITY_TYPE.MULTI_FIELD_PUT:
            self._put(_api, entity, data=properties)
        else:
            for key in field_update_order:
                value = properties.get(key, NOTHING)
                if value is not NOTHING:
                    self._put(_api, entity, data={key : value})
            for key, value in properties.iteritems():
                if key not in field_update_order:
                    self._put(_api, entity, data={key : value})
        if _api.is_aggregating():
            # Result is None when aggregating, since request is sent only later
            return
        if returns is ENTITY:
            return entity
        if returns is MESSAGE:
            return self._result_message(entity, "updated")
        return returns

    @invalidates_autocompletion_cache
    def _rename(self, _api, entity, new_name):
        """ Rename an entity and return a user-friendly message. """
        if entity[self.ENTITY_TYPE.NAME_FIELD] == new_name:
            return self._result_message(entity, 'is already named "{}"'.format(new_name))
        props = {self.ENTITY_TYPE.NAME_FIELD : new_name}
        self._update(_api, entity, returns=ENTITY, **props)
        if _api.is_aggregating():
            # Result is None when aggregating, since request is sent only later
            return
        return self._result_message(entity, 'renamed to "{}"'.format(new_name))

    @invalidates_autocompletion_cache
    def _delete(self, _api, entity_or_entities):
        """
        Delete single or multiple entities and return a user-friendly message.

        :param entity_or_entities: A single entity, or a collection of entities
        """
        entity = self._get_single_entity(entity_or_entities)
        if entity:
            self._delete_entity(_api, entity)
            return self._result_message(entity, "deleted")
        return self._delete_multiple_entities(_api, entity_or_entities)

    def _delete_multiple_entities(self, _api, entities):
        """ Delete multiple entities. """
        with _api.aggregation() as aggregator:
            for entity in entities:
                aggregator.next_task(entity[self.ENTITY_TYPE.NAME_FIELD], "Deleted")
                self._delete_entity(_api, entity)
        return self._multiple_results_message("Deleted {}".format(len(entities)), entities)

    def _delete_entity(self, _api, entity):
        """ Delete a single entity. """
        return _api.delete(self.ENTITY_TYPE.get_uri(entity[self.ENTITY_TYPE.ID_FIELD]))

    def _uri(self, entity, *uri):
        """
        Get URI of collection, collection sub-resource, specific entity, or
        specific entity sub-resource.

        :param entity: Entity for which URI to get. May be ``None``, in which
                       case the collection URI is used
        :param uri: URI parts to append to entity or collection URI
        """
        if entity:
            return self.ENTITY_TYPE.get_uri(entity[self.ENTITY_TYPE.ID_FIELD], *uri)
        else:
            return self.ENTITY_TYPE.get_uri(*uri)

    def _post(self, _api, entity, *uri, **kwargs):
        """
        Shorthand method for entity POST requests. Request URL is computed off
        of the entity's base-URI and the specified URI parts. If ``entity`` is
        ``None``, then the collection URI is used as the base URI.

        :param _api: API object
        :param entity: Entity on which to perform
        :param uri: URI parts to append to the entity's base-URI
        :param kwargs: Optional keyword arguments to delegate to the API instance
        """
        return _api.post(self._uri(entity, *uri), **kwargs)

    def _put(self, _api, entity, *uri, **kwargs):
        """
        Shorthand method for entity PUT requests. Request URL is computed off
        of the entity's base-URI and the specified URI parts. If ``entity`` is
        ``None``, then the collection URI is used as the base URI.

        :param _api: API object
        :param entity: Entity on which to perform
        :param uri: URI parts to append to the entity's base-URI
        :param kwargs: Optional keyword arguments to delegate to the API instance
        """
        return _api.put(self._uri(entity, *uri), **kwargs)

    def _del(self, _api, entity, *uri, **kwargs):
        """
        Shorthand method for entity DELETE requests. Request URL is computed off
        of the entity's base-URI and the specified URI parts.

        :param _api: API object
        :param entity: Entity on which to perform
        :param uri: URI parts to append to the entity's base-URI
        :param kwargs: Optional keyword arguments to delegate to the API instance
        """
        return _api.delete(self.ENTITY_TYPE.get_uri(entity[self.ENTITY_TYPE.ID_FIELD], *uri), **kwargs)

    def _is_single_entity(self, entity_or_entities):
        """
        Test whether the specified parameter is a single entity or a
        collection of entities.

        :param entity_or_entities: A single entity, or a collection of entities
        """
        if isinstance(entity_or_entities, (tuple, list, set)):
            return False
        return isinstance(entity_or_entities, dict) or not hasattr(entity_or_entities, "__iter__")

    def _get_single_entity(self, entity_or_entities):
        """
        If the specified parameter is a single entity, or a collection of
        one entity, return that single entity. Otherwise return ``None``.

        :param entity_or_entities: A single entity, or a collection of entities
        :returns: A single entity, or ``None``.
        """
        if self._is_single_entity(entity_or_entities):
            return entity_or_entities
        if len(entity_or_entities) == 1:
            return entity_or_entities[0]
        return None

    def _perform_single_or_multiple(self, _api, entity_or_entities, func):
        """
        Call the specified function once for each entity.

        :param entity_or_entities: A single entity, or a collection of entities
        :param func: A function, accepting one argument (entity)
        :returns: String representation of all results, separated by new-line
                  characters.
        """
        entities = [entity_or_entities] if self._is_single_entity(entity_or_entities) else entity_or_entities
        results = []
        for entity in entities:
            results.append(func(_api, entity))
        return "\n".join(results)

class EntityFilter(object):
    def __init__(self, **conditions):
        self.conditions = conditions
        self._matchers = None
    def matches(self, entity):
        for property, matcher in self._iter_matchers():
            value = getattr(entity, property, NOTHING)
            if value is NOTHING or not matcher(value):
                return False
        return True
    def matches_all(self, entities):
        return all(self.matches(entity) for entity in entities)
    def filter(self, entities):
        return filter(self.matches, entities)
    def _iter_matchers(self):
        """ Returns iterator of (property, matcher), where the matcher is a
        function that accepts a single entity. """
        if self._matchers is None:
            self._matchers = {property : self._create_matcher(value) for property, value in self.conditions.iteritems()}
        return self._matchers.iteritems()
    def _create_matcher(self, value):
        if isinstance(value, (list, set, tuple)):
            return partial(self.__match_inclusion, set(value))
        return partial(self.__match_equality, value)
    def __match_equality(self, expected, actual):
        return expected == actual
    def __match_inclusion(self, inclusion_list, actual_value):
        return actual_value in inclusion_list

class LazyEntityMapping(object):
    def __init__(self, entity_type, api, fields=None, mapping_key=None, filter=None, preloaded_entities=None):
        self._entity_type = entity_type
        self._mapping_key = mapping_key or entity_type.ID_FIELD
        if not mapping_key and self._is_fetching_by_search() and entity_type.SEARCHABLE:
            func = lambda: entity_type.search(api, page_size=shell_config.api_max_page_size)
        else:
            pluck = None
            if fields and entity_type.PLUCKABLE:
                fields = set((self._mapping_key,)).union(fields)
                pluck = api.create_pluck(*fields)
            func = lambda: entity_type.iter_all(api, pluck=pluck, filter=filter)
        get_mapping = lambda entities: {e[self._mapping_key] : self._get_mapping_value(e) for e in entities if e is not None}  # Management sometimes returns nulls instead of search results
        self._lazy = Lazy(lambda: get_mapping(func()))
        self._preloaded = get_mapping(preloaded_entities or ())
    def get(self, key, default=NOTHING):
        value = self._preloaded.get(key, NOTHING)
        if value is not NOTHING:
            return value
        value = self._lazy.value.get(key, NOTHING)
        if value is not NOTHING:
            return value
        return self._get_default(key)
    def __getitem__(self, key):
        return self.get(key)
    def _is_fetching_by_search(self):
        return False
    def _get_mapping_value(self, entity):
        """ Hook for objects returned by ``get()``. """
        return entity
    def _get_default(self, key):
        """ What's returned when an ID is not mapped. """
        unknown = "UNKNOWN ({}: {})".format(self._mapping_key, key)
        default = NamedDefaultDict(lambda: unknown)
        default[self._mapping_key] = key
        return default

class LazyIdNameMapping(LazyEntityMapping):
    def __init__(self, entity_type, api, filter=None, preloaded_entities=None):
        fields = (entity_type.ID_FIELD, entity_type.NAME_FIELD)
        super(LazyIdNameMapping, self).__init__(entity_type, api, fields=fields, filter=filter, preloaded_entities=preloaded_entities)
    def _is_fetching_by_search(self):
        return True
    def _get_mapping_value(self, entity):
        return entity[self._entity_type.NAME_FIELD]
    def _get_default(self, id):
        return "UNKNOWN (ID: {})".format(id)

def available_to_roles(func=None, roles=None):
    """ Command method availability decorator which filters commands based on
    user roles. This decorator may be used multiple times on the same method.
    If not used, method will be available to all user roles. """
    assert roles, 'Roles must be specified'
    func = availability(func, predicate=lambda: current_system.user.has_role(*roles))
    func = metadata_decorator(func, roles=roles)
    return func

class RolesPredicate(object):
    """
    Convenience wrapper for available_to_roles().
    Usage examples:
        @roles.INFINIDAT.ADMIN.READONLY
        @roles.exclude.READONLY
    """
    __EXCLUDE = "exclude"
    def __init__(self, roles=None, exclude=False):
        self._roles = roles or set()
        self._exclude = exclude
    def __getattr__(self, role):
        # Bridge shell role name to management role name, for backward compatibility
        if role == self.__EXCLUDE:
            assert not self._roles, '"exclude" must be specified at the beginning of the roles list'
            return type(self)(roles=set(Roles.all), exclude=True)
        role_name = getattr(Roles, role.upper())
        new_roles = self._roles.difference([role_name]) if self._exclude else self._roles.union([role_name])
        return type(self)(roles=new_roles)
    def __call__(self, func=None):
        return available_to_roles(func, self._roles)

# Roles predicate "method"
roles = RolesPredicate()

del RolesPredicate

_collection_renderer = CollectionRenderer()
def render_collection(func=None, **kwargs):
    return render(func=func, renderer=_collection_renderer, **kwargs)

def set_rendered_unit(unit):
    if unit is not None:
        set_renderer_properties(
            value_properties=dict(
                capacity_unit=unit,
                ),
            )

def get_renderer_query_fields():
    fields = runtime.options.get("columns")
    if not fields:
        props = get_default_renderer_properties()
        return props.get("fields") or props.get("included") or ()
    return fields

def will_render_any_field(*fields):
    """ Returns ``True`` if any of the specified fields will be rendered for
    the current command, otherwise ``False``. """
    rendered_fields = get_renderer_query_fields()
    return any(f in rendered_fields for f in fields)

def set_renderer_query_fields():
    query_fields = runtime.options.get("columns", ())
    if query_fields:
        if '*' in query_fields:
            columns_option = get_active_shell().current_command.spec.options["columns"]
            if current_system.user.roles.is_infinidat():
                query_fields = columns_option.infinidat_fields or columns_option.fields
            else:
                query_fields = columns_option.fields
        else:
            # Make list unique, keep order
            query_fields = list(OrderedDict.fromkeys(query_fields))
    if is_detailed_query() and query_fields:
        raise BadArguments("--detailed and --columns cannot be specified together")
    if is_detailed_query():
        _set_detailed_query_fields()
    if query_fields:
        _set_specified_query_fields(query_fields)

def _set_specified_query_fields(fields):
    set_renderer_properties(fields=fields)
    defaults = get_default_renderer_properties()
    if "infinidat" in defaults:
        infinidat_props = DictProxy(defaults["infinidat"])
        infinidat_props.update(fields=fields)
        set_renderer_properties(infinidat=infinidat_props)

def _set_detailed_query_fields():
    """ For queries that support the --detailed option, this method sets the
    appropriate rendering properties for detailed mode. This method uses the
    following extra rendering properties:

        detailed_fields - fields to render (overrides ``fields``)
        detailed_substitute - field name substitutions (adds to ``substitute``)
        detailed_individually - whether or not objects are rendered individually
                                in detailed mode
    """
    defaults = get_default_renderer_properties()
    assert "detailed_fields" in defaults
    set_renderer_properties(
        fields=defaults["detailed_fields"],
        substitute=DictProxy(
            defaults.get("substitute", {}),
            defaults.get("detailed_substitute", {}),
            ),
        individually=defaults.get("detailed_individually", defaults.get("individually", False)),
        )
    if "infinidat" in defaults:
        infinidat_props = DictProxy(defaults["infinidat"])
        infinidat_props.update(
            fields=infinidat_props.get("detailed_fields") or infinidat_props.get("fields") or defaults["detailed_fields"],
            substitute=DictProxy(
                defaults.get("substitute", {}),
                infinidat_props.get("substitute", {}),
                defaults.get("detailed_substitute", {}),
                infinidat_props.get("detailed_substitute", {}),
                ),
            individually=infinidat_props.get("detailed_individually", infinidat_props.get("individually", False)),
            )
        set_renderer_properties(infinidat=infinidat_props)
