from munch import Munch
from functools import partial
from collections import defaultdict

from infi.shell import config
from infi.shell.utils import NOTHING
from infi.shell.exceptions import BadArguments

from .types import Pool
from ..exceptions import InvalidSystemState
from . import EntityCommands, LazyIdNameMapping

class TreeNode(object):
    def __init__(self, entity=None, children=None):
        self.entity = entity
        self.children = [] if children is None else children
        self.parent = None
    def add_child(self, treenode):
        self.children.append(treenode)
        treenode.parent = self

class HierarchicalEntityCommandsBase(EntityCommands):
    HIERARCHICAL_TYPE = None  # Subclasses must override this
    
    THIN = "THIN"
    THICK = "THICK"

    PREDICATES = Munch(
        IS_WRITABLE=lambda vol: vol.type != "SNAP",
        )

    def _get_provtype(self, is_thin):
        return self.THIN if is_thin else self.THICK

    def _common_query(self, _api, entities, **params):
        pool_ids = None if not entities or len(entities) > 2 else set(e.pool_id for e in entities)
        return self._query(_api, entities, _hook=self._get_actual_query_hook(_api, pool_ids), **params)

    def _get_actual_query_hook(self, _api, pool_ids=None):
        if pool_ids:
            mapping = {}
            for pool in Pool.iter_all(_api, filter=_api.create_filter().member_of("id", *pool_ids), pluck=_api.create_pluck("id", "name")):
                mapping[pool.id] = pool.name
        else:
            mapping = LazyIdNameMapping(Pool, _api)
        return partial(self._query_hook, pool_names_by_id=mapping)

    def _query_hook(self, entities, pool_names_by_id=None):
        return entities

    def _get_tree(self, _api, entity=None, pool=None):
        if entity and pool:
            raise BadArguments("Either {}, pool, or neither may be specified".format(self.ENTITY_TYPE.get_entity_name()))
        if entity is None:
            roots = self._get_all_trees(_api, pool)
        else:
            roots = self._get_single_tree(_api, entity)
        return self._flatten_tree_for_rendering(roots)

    def _get_all_trees(self, _api, pool=None):
        treenodes = defaultdict(TreeNode)
        roots = []
        for v in self.HIERARCHICAL_TYPE.paged_query(_api,
            hook=self._get_actual_query_hook(_api),
            filter=None if not pool else _api.create_filter(pool_id=pool.id),
            page_size=config.api_max_page_size,
            ):
            treenode = treenodes[v.id]
            treenode.entity = v
            if not v.parent_id:
                roots.append(treenode)
            else:
                treenodes[v.parent_id].add_child(treenode)
        return roots

    def _get_single_tree(self, _api, entity):
        # Apply query hook on single entity
        [entity] = self._get_actual_query_hook(_api)([entity])
        # Compute tree
        root = TreeNode(entity)
        self._fill_single_tree(_api, root)
        return [root]

    def _fill_single_tree(self, _api, parentnode):
        for entity in self.HIERARCHICAL_TYPE.get_all(_api, hook=self._get_actual_query_hook(_api), filter=_api.create_filter(parent_id=parentnode.entity.id)):
            childnode = TreeNode(entity)
            parentnode.add_child(childnode)
            self._fill_single_tree(_api, childnode)

    def _flatten_tree_for_rendering(self, treenodes, depth=0, result=None):
        if result is None:
            result = []
        treenodes.sort(key=lambda treenode: treenode.entity[self.ENTITY_TYPE.NAME_FIELD])
        for treenode in treenodes:
            treenode.entity.name = ('|   ' * depth) + treenode.entity.name
            result.append(treenode.entity)
            self._flatten_tree_for_rendering(treenode.children, depth=depth+1, result=result)
        return result

    def _walk_tree(self, _api, entity):
        for treenode in self._walk_treenodes(self._get_single_tree(_api, entity)):
            yield treenode.entity

    def _walk_treenodes(self, nodes):
        for node in nodes:
            yield node
            for child in self._walk_treenodes(node.children):
                yield child

    def _walk_tree_or_single_entity(self, _api, entity, recurse):
        if recurse:
            return self._walk_tree(_api, entity)
        return (entity,)

    def _update_recursive(self, _api, entity, recurse, success_msg, props, predicate=None):
        if not recurse:
            return self._update(_api, entity, returns=self._result_message(entity, success_msg), **props)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._result_message(entity, success_msg, tree=True),
            )
        single_success_msg = success_msg[0].upper() + success_msg[1:]
        with aggregator:
            for e in self._walk_tree(_api, entity):
                if predicate and not predicate(e):
                    continue
                aggregator.next_task(e.name, single_success_msg)
                self._update(_api, e, returns=NOTHING, **props)
        return aggregator.run()

    def _get_target_pool(self, _api, pool=None):
        """
        Get target pool for creating a new storage entity (volume, fs, etc'). If
        the specified pool is not ``None``, it is returned. Otherwise, if only
        one pool exists in the system, then it is returned. Otherwise, a
        ``BadArguments`` exception is raised.
        """
        if pool:
            return pool
        pools = Pool.paged_query(_api, page_size=1)
        if pools.total_objects == 0:
            raise InvalidSystemState("There are no pools in the system, please create a pool first")
        if pools.total_objects > 1:
            raise BadArguments("There are multiple pools in the system, please specify a pool")
        return next(iter(pools))

    def _result_message(self, entity, msg, subject=None, tree=False):
        return '{} {}{} {}'.format(
            self._result_subject(subject, capitalize=True),
            self._result_entity_name(entity),
            " and all descendants" if tree else "",
            msg,
            )
