import textwrap
import time
import os
import json
from collections import OrderedDict
from functools import partial
from munch import Munch

from infi.shell import (
    commandgroup,
    command,
    render,
    config,
    get_active_shell,
    runtime,
    CommandAborted,
    BadValue,
    )
from infi.shell.types import Str, Int, YesNo
from infi.shell.utils.python import getattr_deep
from infi.shell.utils import strlist
from infi.shell.renderer import set_renderer_properties, get_default_renderer_properties

from ...utils import SafeNestedDict, yesno
from ..exceptions import APIFailure
from ..system import system
from .utils import input_password_check_credentials
from .types import FCPort, SystemService, Node
from .utils import increased_api_timeout
from .options import QueryOptions

from . import roles, EntityCommands

from ..pre_v2_patches import v2_command

class FieldGetter(object):
    def __init__(self, obj, default='N/A'):
        self.obj = obj
        self.default = default
    def __getitem__(self, item):
        return getattr_deep(self.obj, item, self.default)

@commandgroup(prefix="system")
class SystemCommands(object):

    @command
    @render(
        disable_renderer_override=True,
        )
    @roles.INFINIDAT
    def stats_query(self, _api):
        return json.dumps(_api.get("system/stats"), indent=4)

    INFO_SUBSTITUTES = dict(
        rebuild_2_inprogress="rebuild_2_in_progress",
        )

    @command(
        description="Display general system information",
        )
    @render(individually=True)
    def info(self, _api):
        _system = SafeNestedDict(_api.get("system"), default="N/A")
        prop = _system.get_nested
        result = OrderedDict((
            ("name", prop("name")),
            ("serial", prop("serial_number")),
            ("product_id", prop("product_id")),
            ("version", prop("version")),
            ("revision", prop("release.system.revision") if "release" in _system else prop("revision")),
            ("operational_state", prop("operational_state.state")),
            ("wwnn", self._get_system_wwnn(_api)),
            ("total_physical_capacity", prop("capacity.total_physical_capacity")),
            ("free_physical_capacity", prop("capacity.free_physical_space")),
            ("total_virtual_capacity", prop("capacity.total_virtual_capacity")),
            ("free_virtual_capacity", prop("capacity.free_virtual_space")),
            ("ssd_cache_drives", prop("health_state.active_cache_ssd_devices")),
            ))
        for collection in ("pools", "volumes"):
            result[collection] = self._get_total_objects(_api, collection)
        if system.user.roles.is_infinidat():
            for collection in ("filesystems", "exports"):
                result[collection] = self._get_total_objects(_api, collection)
        for collection in ("hosts",):
            result[collection] = self._get_total_objects(_api, collection)
        # Extra info for privileged users
        if system.user.roles.is_infinidat():
            result.update((
                ("deployment_id", prop("deployment_id")),
                ("uptime", prop("uptime")),
                ))
            for field, value in sorted(prop("health_state", {}).iteritems(), key=lambda pair: pair[0]):
                if field in ("active_cache_ssd_devices",):
                    # Some fields are already displayed to all users
                    continue
                if isinstance(value, (dict, list, set, tuple)):
                    continue
                result[self.INFO_SUBSTITUTES.get(field, field)] = value
        return result

    def _get_system_wwnn(self, _api):
        for port in FCPort.iter_all(_api):
            if port.wwnn:
                return port.wwnn
        return None

    def _get_total_objects(self, _api, collection):
        try:
            return _api.get("{}/?page_size=1".format(collection), unwrap_result=False).metadata.number_of_objects
        except APIFailure:
            return "N/A"

    BOOT_PROGRESS_API_TIMEOUT = 5
    BOOT_PROGRESS_REQUEST_INTERVAL = 5
    BOOT_ACTIVATION_TIMEOUT = 120
    @command(
        description="Display system boot progress",
        )
    def boot_progress(self, _api):
        shell = get_active_shell()
        get_info = partial(_api.get, "system/operational_state", timeout=self.BOOT_PROGRESS_API_TIMEOUT)
        erase_line = lambda _line: shell.output("\r" + " "*len(line) + "\r", nl=False, flush=True) if shell.is_interactive_terminal() else shell.output("")
        # Wait for system to respond first
        line = "Waiting for system to respond..."
        shell.output(line, nl=False, flush=True)
        info = None
        while True:
            try:
                info = get_info()
            except APIFailure:
                continue
            break
        erase_line(line)
        # Loop until boot is complete
        while info.state == "STARTING":
            if info.init_state:
                line = "({boot.phase_number}/{boot.total_phases}) {boot.phase}: {boot.description} [{boot.progress_percent}%]".format(boot=info.init_state)
            else:
                line = "State N/A"
            shell.output(line, nl=False, flush=True)
            time.sleep(self.BOOT_PROGRESS_REQUEST_INTERVAL)
            info = get_info()
            erase_line(line)
        if info.state == "STANDBY":
            timeout = time.time() + self.BOOT_ACTIVATION_TIMEOUT
            line = "Waiting for system to complete activation..."
            shell.output(line, nl=False, flush=True)
            while info.state == "STANDBY":
                if time.time() > timeout:
                    erase_line(line)
                    shell.output("Timed out waiting for system to complete activation.")
                    break
                time.sleep(self.BOOT_PROGRESS_REQUEST_INTERVAL)
                info = get_info()
            erase_line(line)
        # Show final system state
        shell.output("System state: {} ({})".format(info.state, info.description), nl=True)

    @command(
        description="Create a system log bookmark for later review by customer support",
        )
    @roles.INFINIDAT
    def log_mark(self, _api):
        _api.internal.post("logs/mark", {})
        return "Logs marked"

    @command(
        description="Save latest system log records for later review by customer support",
        types=(Str, Int)
        )
    @roles.INFINIDAT
    def log_freeze(self, _api, cause=None, duration=None):
        _api.internal.post("logs/freeze", dict(cause=cause, duration=duration))
        return "Logs frozen"

    def _pre_shutdown(self):
        if not (config.auto_confirm_prompts and system.user.roles.is_infinidat()):
            prompt_parts = (
                textwrap.fill("Warning: You are about to shutdown the system. This will stop all I/O services to hosted volumes."),
                textwrap.fill("To avoid potential loss of data, please make sure that no application is accessing the system's volumes. "),
                "Are you sure you wish to continue?",
                )
            if not yesno("\n\n".join(prompt_parts), always_prompt=True):
                raise CommandAborted("Shutdown aborted by user")
            if not input_password_check_credentials("Please enter your password: "):
                raise CommandAborted("Wrong password, shutdown aborted")

    def _shutdown(self, _api, poweroff):
        self._pre_shutdown()
        operation = "shutdown_and_poweroff" if poweroff else "shutdown"
        _api.post("system/operational_state/" + operation)
        msg = "Shutting down system"
        if not poweroff:
            msg += ", keeping power on"
        return msg

    @command(
        name="shutdown",
        description="Initiate system shutdown process",
        types=(YesNo,),
        )
    @roles.INFINIDAT
    def infinidat__shutdown(self, _api, keep_power_on=False):
        return self._shutdown(_api, not keep_power_on)

    @command(
        description="Initiate system shutdown process",
        )
    @roles.ADMIN.TECHNICIAN
    def shutdown(self, _api):
        return self._shutdown(_api, True)

    @command(
        description="Initiate emergency system shutdown process",
        )
    @roles.INFINIDAT
    def emergency_shutdown(self, _api):
        self._pre_shutdown()
        _api.post("system/operational_state/emergency_shutdown")
        return "SYSTEM GOING INTO EMERGENCY SHUTDOWN"

    @command
    @roles.INFINIDAT
    def activate(self, _api):
        _api.put("config/mgmt/state.pause_startup_in", "ACTIVE")
        return "System undergoing activation"

    @command(
        types=(Str,),
        description="Upload an upgrade package",
        )
    @roles.INFINIDAT.ADMIN
    def upgrade(self, _api, filename):
        if os.path.getsize(filename) > 30 * 1024 * 1024:
            raise BadValue("Allowed package size is at most 30 MiB")
        with increased_api_timeout():
            _api.upload("system/upgrade", (filename,))
        return "Package uploaded successfully"



@commandgroup(prefix="service", parent=SystemCommands)
class SystemServiceCommands(EntityCommands):
    ENTITY_TYPE = SystemService
    DETAILED_QUERY_FIELDS = strlist("""
        name
        cluster_state
        """
        )
    QUERY_SUBSTITUTES =dict(
        name="service",
        cluster_state="state",
        )

    @v2_command(
        name="query",
        types=(SystemService.List,),
        description="report state on services",
        substitute=dict(services="name"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=strlist("name cluster_state"),
        substitute=dict(
            name="service",
            cluster_state="state",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def query(self, _api, services=None):
        return self._query(_api, services, _hook=self._query_hook if system.user.roles.is_infinidat() or system.user.has_role("TECHNICIAN") else None)

    def _query_hook(self, services):
        fields = list(get_default_renderer_properties()['fields'])
        fields_set = runtime.options.get("columns", None) is not None
        results = []
        for service in services:
            res = Munch(
                name=service.name,
                cluster_state=service.cluster_state,
                )
            results.append(res)
            states = Munch(("node_{}".format(s.node_id),s) for s in service.node_states)
            for node in sorted(states.keys()):
                res[node] = states[node].role if states[node].state == "ACTIVE" else states[node].state
                if not fields_set:
                    fields.append(node)
            if not fields_set:
                set_renderer_properties(fields=fields)
                fields_set = True
        return results

    @v2_command(
        types=(SystemService, Node),
        description="activate service",
        substitute=dict(service="name"),
        )
    @roles.INFINIDAT.TECHNICIAN
    def activate(self, _api, service, node=None):
        data = dict(node_id=node.id if node else None)
        self._post(_api, service, "start", data=data)
        return self._result_message(service, "commencing activation{}".format("" if not node else " on {}".format(node.name)))

    @v2_command(
        types=(SystemService, Node),
        description="deactivate service",
        substitute=dict(service="name"),
        )
    @roles.INFINIDAT.TECHNICIAN
    def deactivate(self, _api, service, node=None):
        data = dict(node_id=node.id if node else None)
        self._post(_api, service, "stop", data=data)
        return self._result_message(service, "commencing deactivation{}".format("" if not node else " on {}".format(node.name)))
