import inspect
from collections import OrderedDict
from itertools import chain

from infi.shell import (
    OptionType,
    FlagType,
    OptionSet,
    config,
    )
from infi.shell.types import ChoiceType, MappedChoiceType

from ... import debug
from ..renderer import is_csv_mode, set_csv_mode, is_json_mode, set_json_mode
from ..system import system
from .types import Regex

def flag(names, description):
    if isinstance(names, basestring):
        names = (names,)
    class flag(FlagType):
        NAMES = names
        def get_description(self):
            return description
    flag.__name__ = "<Dynamic FlagType: {!r}>".format(names)
    return flag()

class DebugOption(FlagType):
    NAMES = ("d", "xdebug")

    def __init__(self):
        self._changed = False

    def before_command(self, value):
        if value and not debug.is_on():
            self._changed = True
            debug.on()

    def after_command(self, value, result, exc_info):
        if self._changed:
            self._changed = False
            debug.off()

    def is_hidden_from_help(self):
        return True

    def get_description(self):
        return "Show debug information while executing the command."

class AutoApprove(FlagType):
    NAMES = ("y", "yes")

    def __init__(self):
        self._old_value = None

    def before_command(self, value):
        self._old_value = config.auto_confirm_prompts
        config.auto_confirm_prompts = self._old_value or value

    def after_command(self, value, result, exc_info):
        config.auto_confirm_prompts = self._old_value
        self._old_value = None

    def is_hidden_from_help(self):
        return True

    def get_description(self):
        return "Assume yes on all prompts, and do not display them."

class OutputFormat(OptionType):
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            new_cls = type.__new__(mcls, name, bases, attrs)
            # This has to be done here due to cyclic dependencies
            class OutputFormatType(ChoiceType):
                def _get_values(self):
                    return new_cls.get_formats()
            new_cls.ARGUMENT_TYPE = OutputFormatType()
            return new_cls

    NAMES = ("format",)
    ARGUMENT_TYPE = None  # See metaclass

    def __init__(self):
        self._unset_after_command = False

    def before_command(self, format):
        self._unset_after_command = False
        if format is None or self._is_already_set(format):
            return
        self._unset_after_command = True
        self._set(format, True)

    def after_command(self, format, result, exc_info):
        if self._unset_after_command:
            self._set(format, False)

    def _is_already_set(self, format):
        return getattr(self, "_is__" + format)()

    def _set(self, format, enabled):
        getattr(self, "_set__" + format)(enabled)

    @classmethod
    def get_formats(cls):
        if not hasattr(cls, '__output_formats__'):
            cls.__output_formats__ = tuple(attr[5:] for attr in dir(cls)
                                           if attr.startswith('_is__')
                                           and inspect.isroutine(getattr(cls, attr)))
        return cls.__output_formats__

    def _is__csv(self):
        return is_csv_mode()

    def _set__csv(self, enabled):
        set_csv_mode(enabled)

    def _is__json(self):
        return is_json_mode()

    def _set__json(self, enabled):
        set_json_mode(enabled)

class Grep(OptionType):
    NAMES = ("grep",)
    ARGUMENT_TYPE = Regex()

class ExclusiveRegex(Regex):
    @classmethod
    def get_name(cls):
        return Regex.get_name()
    def get_description(self):
        return "A regular expression pattern for exclusion (case insensitive)"

class InvertedGrep(OptionType):
    NAMES = ("inverted-grep", "grepv",)
    ARGUMENT_TYPE = ExclusiveRegex()

class QueryOption(OptionType):

    def __init__(self, fields, infinidat_fields, substitute):
        self.fields = tuple(fields)
        self.infinidat_fields = tuple(infinidat_fields)
        self.substitute = dict(substitute)
        self._arg_type = self._create_argument_type_class(self.fields).List()
        self._infinidat_arg_type = self._arg_type if not infinidat_fields else self._create_argument_type_class(self.infinidat_fields).List()

    def _create_argument_type_class(self, fields):
        raise NotImplementedError

    def get_argument_type(self):
        return self._infinidat_arg_type if system.user.roles.is_infinidat() else self._arg_type

class QuerySort(QueryOption):
    NAMES = ("sort",)

    def _create_argument_type_class(self, fields):
        class Field(MappedChoiceType):
            def _get_mapping(arg_type_self):
                return OrderedDict(chain(
                    ((self.substitute.get(field, field), (field, True)) for field in fields),
                    (("-{}".format(self.substitute.get(field, field)), (field, False)) for field in fields),
                    ))
            def get_fields(arg_type_self):
                return ", ".join(OrderedDict((self.substitute.get(field, field), field) for field in fields))
            def should_render_choices(arg_type_self):
                return False
            def get_description(self):
                return 'Names of sortable columns: e.g, "size" to sort by ascending order, "-size" to sort by descending order. Columns: ' + self.get_fields()
        return Field

class QueryColumns(QueryOption):
    NAMES = ("columns",)

    def _create_argument_type_class(self, fields):
        class Field(MappedChoiceType):
            def _get_mapping(arg_type_self):
                fields_local = fields + ('*',)
                return OrderedDict((self.substitute.get(field, field), field) for field in fields_local)
            def should_render_choices(arg_type_self):
                return False
        return Field

class DetailedQuery(FlagType):
    NAMES = ["detailed"]

    def get_description(self):
        return "Extends the default output with a preset of extra columns."

class QueryOptions(OptionSet):

    def __init__(self, fields, **kwargs):
        """
        Keyword arguments:
          infinidat_fields    - iterable
          sortable            - iterable
          substitute          - dict
          detailed            - boolean
        """
        assert fields, "Must specify at least one field"
        self.fields = tuple(fields)
        self.infinidat_fields = tuple(kwargs.pop("infinidat_fields", ()))
        self.sortable_fields = tuple(kwargs.pop("sortable", ()))
        self.substitute = dict(kwargs.pop("substitute", {}))
        self._detailed = kwargs.pop("detailed", False)
        if kwargs:
            raise TypeError("Unexpected keyword argument(s): " + ", ".join(kwargs))

    def get_options(self):
        sortable_fields = tuple(f for f in self.sortable_fields if f in self.fields)
        infinidat_sortable_fields = tuple(f for f in self.sortable_fields if f in self.infinidat_fields or f in self.fields)
        options = []
        for is_active, cls, args in (
            (self._detailed,                DetailedQuery,      ()),
            (True,                          QueryColumns,       (self.fields, self.infinidat_fields, self.substitute)),
            (bool(self.sortable_fields),    QuerySort,          (sortable_fields, infinidat_sortable_fields, self.substitute)),
            (True,                          OutputFormat,       ()),
            (True,                          Grep,               ()),
            (True,                          InvertedGrep,       ()),
            ):
            if is_active:
                options.append(cls(*args))
        return options
