import itertools
import capacity
from munch import Munch
from functools import partial

from infi.shell import commandgroup, command, get_active_shell
from infi.shell.types import YesNo, Percent
from infi.shell.exceptions import BadArguments
from infi.shell.utils import strlist

from ..system import system
from ..system.user import Roles
from ..renderer.values import StorageCapacityRenderer
from .types import Name, Size, SizeDelta, Pool, AutoExtend, User, CapacityUnit, LocalUser, LdapGroup, PoolAdmin
from .options import QueryOptions
from .utils import are_none
from . import EntityCommands, render_collection, roles, set_rendered_unit

@commandgroup(prefix="pool")
class PoolCommands(EntityCommands):
    ENTITY_TYPE = Pool
    QUERY_FIELDS = strlist("""
        name
        state
        physical_capacity
        allocated_physical_space
        virtual_capacity
        allocated_virtual_space
        max_extend
        """)
    DETAILED_QUERY_FIELDS = strlist("""
        name
        state
        physical_capacity
        allocated_physical_space
        free_physical_space
        virtual_capacity
        allocated_virtual_space
        free_virtual_space
        max_extend
        used_emergency_buffer
        physical_capacity_warning
        physical_capacity_critical
        ssd_enabled
        """)
    QUERY_SUBSTITUTES = dict(
        physical_capacity="physical_total",
        virtual_capacity="virtual_total",
        allocated_physical_space="physical_allocated",
        allocated_virtual_space="virtual_allocated",
        free_physical_space="physical_free",
        free_virtual_space="virtual_free",
        max_extend="emergency_buffer",
        physical_capacity_warning="warning_threshold",
        physical_capacity_critical="critical_threshold",
        ssd_enabled="ssd_cache_default",
        )
    SORTABLE_FIELDS = strlist("""
        name
        physical_capacity
        virtual_capacity
        ssd_enabled
        """)

    @command(
        description="Create a new pool",
        types=(Name, Size, Size, AutoExtend, YesNo),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, physical_capacity, virtual_capacity=None, emergency_buffer=AutoExtend.CONSTANTS.UNLIMITED, ssd_cache=None):
        return self._create(_api,
            name=name,
            virtual_capacity=virtual_capacity or physical_capacity,
            physical_capacity=physical_capacity,
            max_extend=self._get_actual_emergency_buffer_size(name, emergency_buffer),
            ssd_enabled=ssd_cache,
            )

    _EMERGENCY_BUFFER_RESOLUTION = 100*capacity.GiB
    def _get_actual_emergency_buffer_size(self, pool_or_name, size):
        if size is not None and size not in AutoExtend.CONSTANTS.itervalues() and size * capacity.byte % self._EMERGENCY_BUFFER_RESOLUTION != 0:
            size += (self._EMERGENCY_BUFFER_RESOLUTION - size * capacity.byte % self._EMERGENCY_BUFFER_RESOLUTION) / capacity.byte
            pool = Munch(name=pool_or_name) if isinstance(pool_or_name, basestring) else pool_or_name
            render_capacity = partial(StorageCapacityRenderer().render, empty_value="N/A")
            get_active_shell().add_warning("Emergency buffer for {} was rounded up to {} ({})".format(
                self._result_subject_and_entity(pool), render_capacity(size, binary_capacity=True), render_capacity(size)))
        return size

    @command(
        description="List existing pools",
        types=(Pool.List, CapacityUnit),
        substitute=dict(pool="name"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Pool,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, pool=None, unit=None):
        set_rendered_unit(unit)
        return self._query(_api, pool, _hook=self._query_hook)

    def _query_hook(self, pools):
        for p in pools:
            p.max_extend = "UNLIMITED" if p.max_extend < 0 else "DISABLED" if not p.max_extend else p.max_extend
            p.allocated_virtual_space = p.virtual_capacity - p.free_virtual_space
            used_delta = p.allocated_physical_space - p.physical_capacity
            p.used_emergency_buffer = None if used_delta < 0 else used_delta
            for field in ("physical_capacity_warning", "physical_capacity_critical"):
                p[field] = "{}%".format(p[field])
        return pools

    @command(
        description="Grant pool provisioning privileges to a pool administrator",
        types=(Pool, PoolAdmin.List),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def add_admin(self, _api, pool, admin):
        with _api.aggregation() as aggregator:
            for user in admin:
                aggregator.next_task(self._get_user_subject(user), "Granted provisioning privileges on pool {}.".format(pool.name))
                _api.post(Pool.get_uri(pool.id, "owners", user.id))
        return self._multiple_results_message("Granted provisioning privileges on pool {} to".format(pool.name), [self._get_user_subject(user) for user in admin], PoolAdmin)
    @command(
        description="Revoke pool provisioning privileges from a pool administrator",
        types=(Pool, PoolAdmin.List),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove_admin(self, _api, pool,admin):
        with _api.aggregation() as aggregator:
            for user in admin:
                aggregator.next_task(self._get_user_subject(user), "Revoked provisioning privileges on pool {}.".format(pool.name))
                _api.delete(Pool.get_uri(pool.id, "owners", user.id))
        return self._multiple_results_message("Revoked provisioning privileges on pool {} from".format(pool.name), [self._get_user_subject(user) for user in admin], PoolAdmin)

    def _get_user_subject(self, user):
        return '{} "{}"'.format((LocalUser if user.type.lower() == "local" else LdapGroup).get_entity_name(), user.name)

    @command(
        description="Rename pool",
        types=(Pool, Name),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, pool, new_name):
        return self._rename(_api, pool, new_name)

    @command(
        description="Resize pool",
        types=(Pool, SizeDelta, SizeDelta, AutoExtend),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def resize(self, _api, pool, physical_capacity=None, virtual_capacity=None, emergency_buffer=None):
        if not any(arg is not None for arg in (physical_capacity, virtual_capacity, emergency_buffer)):
            raise BadArguments("At least one more argument must be specified")
        emergency_buffer = self._get_actual_emergency_buffer_size(pool, emergency_buffer)
        if physical_capacity is not None:
            physical_capacity = physical_capacity.compute_size(pool.physical_capacity)
        if virtual_capacity is not None:
            virtual_capacity = virtual_capacity.compute_size(pool.virtual_capacity)
        # The following lines should be erased once management supports multi-field
        # updates, and the next commented lines should be uncommented in their place:
        self._update_attrs(_api, pool,
            ("Physical capacity", "physical_capacity", physical_capacity),
            ("Virtual capacity", "virtual_capacity", virtual_capacity),
            ("Max extend", "max_extend", emergency_buffer),
            )
        # self._update(_api, pool,
        #     virtual_capacity=virtual_capacity,
        #     physical_capacity=physical_capacity,
        #     max_extend=emergency_buffer,
        #     )
        return self._result_message(pool, "resized" if emergency_buffer is None
                                          else "emergency-buffer policy updated" if are_none(physical_capacity, virtual_capacity)
                                          else "resized, and its mergency-buffer policy updated")

    @command(
        description="Set pool capacity notification thresholds",
        types=(Pool, Percent, Percent),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thresholds(self, _api, pool, warning, critical):
        warning_changed = warning != pool.physical_capacity_warning
        critical_changed = critical != pool.physical_capacity_critical
        if not any((warning_changed, critical_changed)):
            return self._result_message(pool, "warning threshold is already {}%, critical threshold is already {}%".format(warning, critical))
        # The following lines should be erased once management supports multi-field
        # updates, and the next commented lines should be uncommented in their place:
        actions = (
            ("Warning threshold", "physical_capacity_warning", warning),
            ("Critical threshold", "physical_capacity_critical", critical),
            )
        if critical > pool.physical_capacity_critical:
            # If critical is increased, it should be changed first
            actions = reversed(actions)
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for task, attr, value in actions:
                aggregator.next_task(task, "updated")
                self._update(_api, pool, **{attr : value})
        aggregator.run()
        # self._update(_api, pool
        #     physical_capacity_warning=warning,
        #     physical_capacity_critical=critical,
        #     )
        template_msg = "{} threshold changed from {}% to {}%"
        warn_msg = template_msg.format("warning", pool.physical_capacity_warning, warning) if warning_changed else None
        crit_msg = template_msg.format("critical", pool.physical_capacity_critical, critical) if critical_changed else None
        return self._result_message(pool, ", ".join(m for m in (warn_msg, crit_msg) if m))

    def _update_attrs(self, _api, pool, *attr_tuples, **kwargs):
        """ Keyword arguments: success_msg """
        aggregator = _api.create_request_aggregator(
            total_success_msg=kwargs.pop("success_msg", None) or self._result_message(pool, "updated"),
            )
        with aggregator:
            for task, attr, value in attr_tuples:
                aggregator.next_task(task, "updated")
                self._update(_api, pool, **{attr : value})
        return aggregator.run()

    # @command(
    #     description="Move all volumes from one pool to another, and delete the first pool",
    #     types=(Pool, Pool),
    #     substitute=dict(
    #         source="name",
    #         target="into",
    #         ),
    #     )
    # @roles.INFINIDAT.ADMIN
    # def merge(self, _api, source, target):
    #     raise NotImplementedError("No support on server side")

    @command(
        description="Lock pool, revoking write access to its volumes",
        types=(Pool,),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def lock(self, _api, pool):
        self._post(_api, pool, "lock")
        return self._result_message(pool, "locked")

    @command(
        description="Unlock pool, granting write access to its volumes",
        types=(Pool,),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def unlock(self, _api, pool):
        self._post(_api, pool, "unlock")
        return self._result_message(pool, "unlocked")

    @command(
        description="Enable or disable SSD cache",
        types=(Pool, YesNo,),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def cache(self, _api, pool, ssd):
        action_name = "enabled" if ssd else "disabled"
        if pool.ssd_enabled == ssd:
            return self._result_message(pool, "SSD cache is already " + action_name)
        self._update(_api, pool,
            ssd_enabled=ssd,
            )
        return self._result_message(pool, "SSD cache " + action_name)

    @command(
        description="Delete pool",
        types=(Pool.List,),
        substitute=dict(pool="name"),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, pool):
        return self._delete(_api, pool)

    @command(
        description="List existing pools with respective administrators",
        types=(Pool,),
        substitute=dict(pool="name"),
        )
    def user_query(self, _api, pool=None):
        users = {u.id : u for u in User.iter_all(_api)}
        default_admin_roles = set((Roles.INFINIDAT, Roles.ADMIN)) if system.user.roles.is_infinidat() else set((Roles.ADMIN,))
        default_admins = sorted(u.name for u in users.itervalues() if any(role.upper() in default_admin_roles for role in u.roles))
        return (Munch(
                    name=p.name,
                    admins=sorted(itertools.chain((users[uid].name for uid in p.owners), default_admins)),
                    )
                for p in ([pool] if pool else self._query(_api)))
