from infi.shell.utils import ContextProxy

from ...utils import Proxy
from ..exceptions import Unauthorized, Forbidden, ManagementException, UnexpectedAPIResponse
from .user import User

system = ContextProxy('system context')
api = Proxy(lambda: system.api)

class System(object):
    def __init__(self, api):
        self.api = api
        self._user = None
        self._system_name = None
    def name(self):
        if self._system_name is None:
            self._system_name = self.api.get('system/name')
        return self._system_name
    def _login_and_get_user(self, username, password):
        with self.api.no_basic_auth_context():
            return self.api.post("users/login", dict(username=username, password=password))
    def check_credentials(self, username, password):
        """ Attempt to login, but do not change the current user. Return True if
        login succeeded, else False. """
        try:
            self._login_and_get_user(username, password)
        except (Unauthorized, Forbidden):
            return False
        except ManagementException as e:
            if e.http_code in (401, 403):
                return False
            raise
        return True
    def login(self, username, password):
        principal = self._login_and_get_user(username, password)
        self._assert_principal(principal)
        self._set_user(principal.name, principal.roles)
        self.api.set_basic_auth(username, password)
    def _assert_principal(self, principal):
        if not principal:
            raise UnexpectedAPIResponse("User principal is empty")
        for field in "name roles".split():
            if field not in principal:
                raise UnexpectedAPIResponse("Missing field in user principal: " + field)
    def _set_user(self, *args, **kwargs):
        from ..commands.types import ConstantEntity  # Avoid cyclic imports
        self._user = User(*args, **kwargs)
        ConstantEntity.invalidate_all_caches()
    def logout(self):
        self._user = None
    def is_logged_in(self):
        return self._user is not None
    @property
    def user(self):
        assert self.is_logged_in(), 'User not set - invoke login(username, password)'
        return self._user
    def __enter__(self):
        system.__push_context__(self)
    def __exit__(self, exc_type, exc_instance, traceback):
        system.__pop_context__(self)
