import sys
import re
from itertools import chain
from collections import defaultdict, OrderedDict
from functools import wraps

from infi.shell import config, get_active_shell
from infi.shell.utils import input as input_module, NOTHING
from infi.shell.rlcontext import get_raw_input_functions
from infi.shell.exceptions import CommandAborted

class PossiblyNone(object):
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            base_methods = 'add sub mul div truediv floordiv mod divmod pow lshift rshift and xor or'.split()
            rbase_methods = ['r'+m for m in base_methods]
            ibase_methods = ['i'+m for m in base_methods]
            other_methods = 'neg pos abs invert complex int long float oct hex'.split()
            methods = chain(base_methods, rbase_methods, ibase_methods, other_methods)

            underscores = lambda m: '__{}__'.format(m)

            def create_operator(method_name): # This is a method because method_name must not change
                def operator(self, *args):
                    if self._is_none:
                        return self
                    args = list(args)
                    invoked_operand = self._value
                    actual_method_name = method_name
                    if method_name in base_methods and args and isinstance(args[0], type(self)):
                        actual_method_name = 'r' + actual_method_name
                        invoked_operand, args[0] = args[0], invoked_operand
                    actual_method_name = underscores(actual_method_name)
                    real_op_method = getattr(invoked_operand, actual_method_name, None)
                    if real_op_method is None:
                        raise NotImplementedError
                    return real_op_method(*args)
                operator.__name__ = method_name
                return operator

            for op_method_name in methods:
                attrs[underscores(op_method_name)] = create_operator(op_method_name)
            return type.__new__(mcls, name, bases, attrs)

    def __init__(self, value, alternative_value='N/A'):
        self._is_none = value is None
        self._value = value
        self._alternative_value = unicode(alternative_value)
        def __setattr__(self, attr, value):
            if not self._is_none:
                setattr(self._value, attr, value)
        self.__setattr__ = __setattr__
    def __repr__(self):
        return '<PossiblyNone: {}>'.format(repr(self._value))
    def __str__(self):
        return self._alternative_value if self._is_none else str(self._value)
    def __unicode__(self):
        return self._alternative_value if self._is_none else unicode(self._value)
    def __nonzero__(self):
        return True if self._is_none else bool(self._value)
    def __getattr__(self, attr):
        return self if self._is_none else getattr(self._value, attr)
    def __hasattr__(self, attr):
        return False if self._is_none else hasattr(self._value, attr)
    def __getitem__(self, item):
        return self if self._is_none else self._value[item]
    def __setitem__(self, item, value):
        if not self._is_none:
            self._value[item] = value
    def __delitem__(self, item):
        if not self._is_none:
            del self._value[item]
    def __len__(self):
        return 0 if self._is_none else len(self._value)
    def __iter__(self):
        return iter(()) if self._is_none else iter(self._value)
    def __reversed__(self):
        return self if self._is_none else reversed(self._value)
    def __contains__(self, value):
        return False if self._is_none else value in self._value

    @classmethod
    def wrap_items(cls, container, **kwargs):
        if isinstance(container, dict):
            return type(container)((key, cls.wrap_items(value, **kwargs)) for key, value in container.iteritems())
        elif isinstance(container, (list, tuple, set)):
            return type(container)(cls.wrap_items(value, **kwargs) for value in container)
        else:
            # Container is actually the value here
            return cls(container, **kwargs)

class NamedDict(dict):
    __nothing__ = object()
    def __init__(self, *args, **kwargs):
        self.__dict_name__ = kwargs.pop('__dict_name__', type(self).__name__)
        dict.__init__(self, *args, **kwargs)
    def __getattr__(self, attr):
        value = self.get(attr, self.__nothing__)
        if value is self.__nothing__:
            raise AttributeError('{} has no such attribute: {}'.format(self.__dict_name__, attr))
        return value

class NamedDefaultDict(defaultdict):
    def __getattr__(self, attr):
        return self[attr]

class NamedOrderedDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        super(NamedOrderedDict, self).__init__(*args, **kwargs)
        def __getattr__(self, attr):
            return self[attr]
        self.__getattr__ = __getattr__
    @classmethod
    def constants(cls, *values):
        return cls((v if isinstance(v, (tuple, list)) else (v, v) for v in values))

class Proxy(object):
    __allowed_attrs__ = set(('__getter_func__',))

    def __init__(self, getter_func):
        self.__getter_func__ = getter_func

    def __getattr__(self, attr):
        return getattr(self.__getter_func__(), attr)

    def __setattr__(self, attr, value):
        if attr in self.__allowed_attrs__:
            return object.__setattr__(self, attr, value)
        setattr(self.__getter_func__(), attr, value)

    def __getitem__(self, item):
        return self.__getter_func__()[item]

    def __setitem__(self, item, value):
        self.__getter_func__()[item] = value

class SafeNestedDict(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args)
        self._default = kwargs.pop("default", None)
        if kwargs: raise TypeError("Unexpected keyword arguments: " + ", ".join(kwargs))
    def get_nested(self, nested_item, default=NOTHING, replace_null=False):
        if default is NOTHING:
            default = self._default
        result = self
        for item in nested_item.split("."):
            if not isinstance(result, dict):
                return default
            result = result.get(item, NOTHING)
            if result is NOTHING:
                return default
        if result is None and replace_null:
            return default
        return result

class Version(object):
    _pattern = re.compile(r"^(\d+(?:\.\d+)*)-?(.*)$")
    def __init__(self, version_string):
        self.text = str(version_string) if version_string else "0"
        self.parts, self.suffix = self.parse(self.text)
        self._is_zero = not self.suffix and not any(self.parts)  # No suffix and all parts are zero
    @classmethod
    def parse(cls, version_string):
        match = cls._pattern.match(version_string)
        parts, suffix = (0,), ''
        if match:
            groups = match.groups()
            if groups:
                parts = tuple(int(num) for num in groups[0].split("."))
                if len(groups) > 1:
                    suffix = groups[1]
        return parts, suffix
    def _compare(self, v):
        if len(self.parts) == len(v.parts):
            return 0 if self.parts == v.parts else -1 if self.parts < v.parts else 1
        common_len = min(len(self.parts), len(v.parts))
        if len(self.parts) > len(v.parts):
            longer, diff = self.parts, 1
            my_common, other_common = self.parts[:common_len], v.parts
        else:
            longer, diff = v.parts, -1
            my_common, other_common = self.parts, v.parts[:common_len]
        if my_common == other_common:
            for part in longer[common_len:]:
                if part > 0:
                    return diff
            return 0
        return -1 if my_common < other_common else 1
    def __eq__(self, v): return self._compare(v) == 0
    def __ne__(self, v): return self._compare(v) != 0
    def __lt__(self, v): return self._compare(v) <  0
    def __le__(self, v): return self._compare(v) <= 0
    def __gt__(self, v): return self._compare(v) >  0
    def __ge__(self, v): return self._compare(v) >= 0
    def __nonzero__(self):
        return not self._is_zero
    def __str__(self):
        return self.text
    def __repr__(self):
        return "Version({!r})".format(self.text)

def humanly_join(strings):
    if not isinstance(strings, (list, tuple)):
        strings = tuple(strings)
    if len(strings) == 1:
        return strings[0]
    return "{} and {}".format(", ".join(strings[:-1]), strings[-1])

def capitalize(s):
    return '' if not s else (s[0].capitalize() + s[1:])

INPUT_INTERACTION_ERROR = "Command requires user interaction, please use the interactive shell"

def _ensure_interactive_shell_for_input(input_kwargs):
    interaction_error = input_kwargs.pop("interaction_error", INPUT_INTERACTION_ERROR)
    interaction_exc_type = input_kwargs.pop("interaction_exc_type", CommandAborted)
    if not get_active_shell().is_interactive_terminal():
        raise interaction_exc_type(interaction_error)

def _wrap_input_func(func):
    """
    Wrapper for input functions from ``infi.shell.utils.input``, that first
    ensures the shell is in interactive mode. The wrapped functions support
    additional parameters:

    ======================    ================================================
    Parameter                 Meaning
    ======================    ================================================
    interaction_error         words to add to error message in non-interactive
                              mode
    interaction_exc_type      exception to raise in non-interactive mode
                              (defaults to ``CommandAborted``)
    ======================    ================================================
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        _ensure_interactive_shell_for_input(kwargs)
        return func(*args, **kwargs)
    return wrapper

input_text               = _wrap_input_func(input_module.input_text)
input_password           = _wrap_input_func(input_module.input_password)
input_choice             = _wrap_input_func(input_module.input_choice)
input_simple_choice      = _wrap_input_func(input_module.input_simple_choice)

def yesno_ensure_interactive(always_prompt=False):
    """
    Validation for ``yesno()``

    :param always_prompt: Ignore config.auto_confirm_prompts
    :returns: ``True`` if user should NOT be prompted, meaning the answer is YES; otherwise ``False``.
    """
    if config.auto_confirm_prompts and not always_prompt:
        return True
    if not get_active_shell().is_interactive_terminal():
        if always_prompt:
            raise CommandAborted("Command requires user confirmation (use the interactive shell)")
        else:
            raise CommandAborted("Command requires user confirmation (use -y or the interactive shell)")
    return False

def yesno(*args, **kwargs):
    """
    Ensure shell is interactive, and delegate to ``infi.shell.utils.input.yesno``.

    :param prompt: Prompt to display to user.
    :param always_prompt: Ignore config.auto_confirm_prompts
    :returns: ``True`` for "yes", ``False`` for "no".
    """
    always_prompt = kwargs.pop('always_prompt', False)
    if yesno_ensure_interactive(always_prompt):
        return True
    return input_module.yesno(*args, **kwargs)

@_wrap_input_func
def get_username_password(username=None, prompt=None):
    raw_input, getpass = get_raw_input_functions()
    prompt = "{}: ".format(prompt or "Username")
    while username is None:
        username = raw_input(prompt)
    password = getpass()
    return username, password

def post_mortem(exc_info=None):
    exc_info = exc_info or sys.exc_info()
    try:
        import ipdb as pdb
    except ImportError:
        import pdb # noqa
    print("\n### POST MORTEM ###")
    pdb.post_mortem(exc_info[2])

def merge_dicts(*dicts):
    result = {}
    for d in dicts:
        result.update(d)
    return result
